package com.simonstl.gorille;


/*
The contents of this file are subject to the Mozilla Public License  Version 1.1 (the "License"); you may not use this file except in  compliance with the License. You may obtain a copy of the License at 
http://www.mozilla.org/MPL/ 
Software distributed under the License is distributed on an "AS IS"  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  License for the specific language governing rights and limitations under  the License. 

The Original Code is available at  http://simonstl.com/projects/fragment/original. 

The Initial Developer of the Original Code is Simon St.Laurent. Portions  created by Simon St.Laurent are Copyright (C) 2001. All Rights Reserved. 

Contributor(s): 
*/

/**
* <p>This class stores a set of ranges of characters, allowing XML processors to test against different kinds of ranges.</p>
*
*<p>Version 0.07 fixes the surrogate pairs code so that it returns false if the second code value in a surrogate pair isn't a low surrogate.</p>
*<p>Version 0.06 fixes the character-checking so it rejects low surrogates that appear without a prior high surrogate.  Thanks to Elliotte Rusty Harold for pointing this out.</p>
*
*<p>Version 0.04 changes from char values to int values in order to support Unicode characters (surrogate pairs) outside of Plane 0.  Thanks to Elliotte Rusty Harold for noticing, and John Cowan for suggesting a solution.</p>
* <pre>for each char in the String
	if in the range 0xD800 to 0xDBFF
		grab the next char as well
		set int[] element to
		(this-0xD800) * 0x400 + (next-0xDC00) + 10000
	else
		set int[] element to value of this char
</pre>
* Version 0.03 adds additional support for namespace-handling.
* Version 0.02 adds support for pubidChar.
*
* @version 0.05 10 January 2002
* @author Simon St.Laurent
*/

public class CharRules {

protected CharRanges chars;
protected CharRanges nameStartChar;
protected CharRanges addlNameChar;
protected CharRanges space;
protected CharRanges baseChar;
protected CharRanges ideographic;
protected CharRanges combiningChar;
protected CharRanges digit;
protected CharRanges extender;
protected CharRanges pubidChar;
protected boolean namespaceAware=true;

public CharRules(){
}

public void setNamespaceAware(boolean _namespaceAware) {
  namespaceAware=_namespaceAware;
}

public boolean getNamespaceAware(){
  return namespaceAware;
}

/** Sets the list of overall characters for content.  (The Char production in XML 1.0 and 1.1.) **/

public void setChar(CharRanges _chars){
  chars=_chars;
}


/** Sets the range representing the list of characters for use at the beginning of names.  (The NameStartChar production in XML 1.1.) **/

public void setNameStartChar(CharRanges _nameStartChar){
  nameStartChar=_nameStartChar;
}

/** Sets the range representing the list of additional characters which can be used in names after they start.  (The NameChar production in XML 1.1, minus the NameStartChar.)**/

public void setAddlNameChar(CharRanges _addlNameChar){
  addlNameChar=_addlNameChar;
}

/** Sets the range representing the characters which represent whitespace. (The S production in XML 1.0 and 1.1.)**/
public void setSpace(CharRanges _space){
  space=_space;
}

/** Sets the range representing the BaseChar production from XML 1.0. **/
public void setBaseChar(CharRanges _baseChar){
  baseChar=_baseChar;
}

/** Sets the range representing the Ideographic production from XML 1.0**/
public void setIdeographic(CharRanges _ideographic){
  ideographic=_ideographic;
}

/** Sets the range representing the CombiningChar production from XML 1.0**/
public void setCombiningChar(CharRanges _combiningChar){
  combiningChar=_combiningChar;
}

/** Sets the range representing the Digit production from XML 1.0**/
public void setDigit(CharRanges _digit){
  digit=_digit;
}

/** Sets the range representing the Extender production from XML 1.0**/
public void setExtender(CharRanges _extender){
  extender=_extender;
}

/** Sets the range representing the PubidChar production from XML 1.0**/
public void setPubidChar(CharRanges _pubidChar){
  pubidChar=_pubidChar;
}

/** Checks characters against the range of acceptable Char characters.**/

public boolean isChar(int test){
  boolean returnVal=false;
  if (chars != null) {
    returnVal=chars.isInRanges(test);
  }
  return returnVal;
}


/** Checks characters at the beginning of names against the range of acceptable NameStartChar characters, if there is one, and falls back on XML 1.0's Letter rules if there isn't.**/

public boolean isNameStartChar(int test){
  boolean returnVal=false;
  if (nameStartChar != null) {
    returnVal=nameStartChar.isInRanges(test);
  } else {
    //fall back on XML 1.0 if no nameStartChar
    returnVal=this.isLetter(test);
  }

  return returnVal;
}

/** Checks characters after the beginning of names against the range of acceptable additional name characters, if there is one.  **/

public boolean isAddlNameChar(int test){
  boolean returnVal=false;
  if (addlNameChar != null) {
    returnVal=addlNameChar.isInRanges(test);
  }
  return returnVal;
}

/** Checks characters to see if they are whitespace.**/

public boolean isSpace(int test){
  boolean returnVal=false;
  if (space != null) {
    returnVal=space.isInRanges(test);
  }
  return returnVal;

}

/** Checks characters to see if they are part of the BaseChar range.**/

public boolean isBaseChar(int test){
  boolean returnVal=false;
  if (baseChar != null) {
    returnVal=baseChar.isInRanges(test);
  }
  return returnVal;
}
/** Checks characters to see if they are part of the Ideographic range.**/

public boolean isIdeographic(int test){
  boolean returnVal=false;
  if (ideographic != null) {
    returnVal=ideographic.isInRanges(test);
  }
  return returnVal;

}
/** Checks characters to see if they are part of the CombiningChar range.**/

public boolean isCombiningChar(int test){
  boolean returnVal=false;
  if (combiningChar != null) {
    returnVal=combiningChar.isInRanges(test);
  }
  return returnVal;

}
/** Checks characters to see if they are part of the Digit range.**/

public boolean isDigit(int test){
  boolean returnVal=false;
  if (digit != null) {
    returnVal=digit.isInRanges(test);
  }
  return returnVal;

}
/** Checks characters to see if they are part of the Extender range.**/

public boolean isExtender(int test){
  boolean returnVal=false;
  if (extender != null) {
    returnVal=extender.isInRanges(test);
  }
  return returnVal;

}

/** Checks characters to see if they are part of the PubidChar range.**/

public boolean isPubidChar(int test){
  boolean returnVal=false;
  if (pubidChar != null) {
    returnVal=pubidChar.isInRanges(test);
  }
  return returnVal;

}


/** Checks characters to see if they are part of the Letter range - in XML 1.0, that's the combination of BaseChar and Ideographic.**/

public boolean isLetter(int test){
  boolean returnVal=this.isBaseChar(test);
   if (!returnVal) {
     returnVal=this.isIdeographic(test);
   }
  return returnVal;
}

/** Checks characters to see if they are part of the Name range - in XML 1.0, that's the combination of BaseChar, Ideographic, CombiningChar, and Extender.  In XML 1.1, that's NameStartChar and the additional name characters.**/

public boolean isNameChar(int test){
  boolean returnVal=false;

if ((namespaceAware) && (test==':')) {
  returnVal=false;
} else {
  if (!((nameStartChar==null) && (addlNameChar==null))){
    returnVal=this.isNameStartChar(test);
    if (!returnVal){
      returnVal=this.isAddlNameChar(test);
    }

   } else {
   //fall back on XML 1.0 rules

   if (test=='.') {returnVal=true;}
   if (!returnVal) {
     if (test=='-') {returnVal=true;}
   }
   if (!returnVal) {
     if (test=='_') {returnVal=true;}
   }
   if (!returnVal) {
     if (test==':') {returnVal=true;}
   }
   if (!returnVal) {
     returnVal=this.isLetter(test);
   }
   if (!returnVal) {
     returnVal=this.isDigit(test);
   }
   if (!returnVal) {
     returnVal=this.isCombiningChar(test);
   }
   if (!returnVal) {
     returnVal=this.isExtender(test);
   }
 }
}
  return returnVal;
}


/** Checks a String to see if it conforms to the Chars rules. Works using either XML 1.0 or 1.1, depending on the rules file used to set the ranges. **/
public boolean isChars(String test) {
  boolean returnVal=false;
  int length=test.length();

  if (length>0) {
     for (int i=0; i<length; i++) {
        int testChar=test.charAt(i);
        if ((testChar>=0xDC00) && (testChar<=0xDFFF)){ return false; }
        if ((testChar>=0xD800) && (testChar<=0XDBFF)){
          if (i<(length-1)) {
            int secondChar=test.charAt(++i);
            if ( (secondChar<0xDC00) || (secondChar>0xDFFFF) ) { return false; }
            int newChar=( ((testChar-0xD800) * 0x400) + (secondChar-0xDC00) + 0x10000);

            testChar=newChar;
          } //end length
         } 

        returnVal=this.isChar(testChar);
        if (!returnVal) {break;}
     }//end for
  }//end length>0
  return returnVal;
}

/** Checks a character array from SAX to see if it conforms to the Chars rules. Works using either XML 1.0 or 1.1, depending on the rules file used to set the ranges. **/
public boolean isChars(char ch[], int start, int length) {
  boolean returnVal=false;
//System.out.println("Start: "+start+" Length: "+length);
     for (int i=start; i<start+length; i++) {
        int testChar=ch[i];
        if ((testChar>=0xDC00) && (testChar<=0xDFFF)){ return false; }
        if ((testChar>=0xD800) && (testChar<=0XDBFF)){
          if (i<(length-1)) {
           int secondChar=ch[++i];
           if ( (secondChar<0xDC00) || (secondChar>0xDFFFF) ) { return false; }

           int newChar=( ((testChar-0xD800) * 0x400) + (secondChar-0xDC00) + 0x10000);

           testChar=newChar;
          }
         }

        returnVal=this.isChar(testChar);
//System.out.println("Value: "+testChar+" result: "+returnVal);
        if (!returnVal) {break;}
     }//end for
  return returnVal;
}


/** Checks a String to see if it conforms to the Name rules. Works using either XML 1.0 or 1.1, depending on the rules file used to set the ranges. **/
public boolean isName(String test) {
  boolean returnVal=false;
  int length=test.length();
  int start=1;

  if (length>0) {
    int firstChar=test.charAt(0);
//System.out.println("Testing: "+firstChar);
    if ((firstChar>=0xDC00) && (firstChar<=0xDFFF)){ return false; }
    if ((firstChar>=0xD800) && (firstChar<=0XDBFF)){
      if (length>1) {
       int secondChar=test.charAt(1);
//System.out.println("Testing: "+secondChar);

       if ( (secondChar<0xDC00) || (secondChar>0xDFFFF) ) { return false; }

       int newChar=( ((firstChar-0xD800) * 0x400) + (secondChar-0xDC00) + 0x10000);

       firstChar=newChar;
//System.out.println("Final:"+firstChar);
       start++;
       }
    } 

   returnVal=this.isNameStartChar(firstChar);
//System.out.println("firstChar result"+returnVal);
   if (returnVal && (length>1)) { //first char passed
      for (int i=start; i<length; i++) {
        int testChar=test.charAt(i);
        if ((testChar>=0xDC00) && (firstChar<=0xDFFF)){ return false; }
        if ((testChar>=0xD800) && (testChar<=0XDBFF)){
         if (i<(length-1)) {

          int secondChar=test.charAt(++i);
          if ( (secondChar<0xDC00) || (secondChar>0xDFFFF) ) { return false; }

          int newChar=( ((testChar-0xD800) * 0x400) + (secondChar-0xDC00) + 0x10000);

          testChar=newChar;
         }
        }
        returnVal=this.isNameChar(testChar);
        if (!returnVal) {break;}
      }//end for
    }//end if
  }//end length>0
  return returnVal;
}

/** Checks a String to see if it conforms to the QName rules. Works using either XML 1.0 or 1.1, depending on the rules file used to set the ranges. **/
public boolean isQName(String test) {
  boolean returnVal=false;
  int length=test.length();
  int numColons=0;
  int start=1;

  if (length>0) {
    int firstChar=test.charAt(0);
    if ((firstChar>=0xDC00) && (firstChar<=0xDFFF)){ return false; }
    if ((firstChar>=0xD800) && (firstChar<=0xDBFF)){
     if (length>1) {
      int secondChar=test.charAt(1);
      if ( (secondChar<0xDC00) || (secondChar>0xDFFFF) ) { return false; }

      int newChar=( ((firstChar-0xD800) * 0x400) + (secondChar-0xDC00) + 0x10000);

      firstChar=newChar;
      start++;
     } 
    } 

    returnVal=this.isNameStartChar(firstChar);

    if (returnVal && (length>1)) { //first char passed
      for (int i=start; i<length; i++) {
        int testChar=test.charAt(i);
        if ((testChar>=0xDC00) && (testChar<=0xDFFF)){ return false; }
        if ((testChar>=0xD800) && (testChar<=0XDBFF)){
         if (i<(length-1)) {

          int secondChar=test.charAt(++i);
          if ( (secondChar<0xDC00) || (secondChar>0xDFFFF) ) { return false; }

          int newChar=( ((testChar-0xD800) * 0x400) + (secondChar-0xDC00) + 0x10000);

          testChar=newChar;
         } 
        } 
        returnVal=this.isNameChar(testChar);
        if (!returnVal) {
          if (testChar!=':') {
            break; //bad char not a colon
          } else {
            numColons++;
            if ((numColons>1)||(i==(length-1))){
              break; //too many colons, or colon at end
            }
           //okay, it's a name with one colon and a character following. 
           //next char must be a namestart
           i++;
           testChar=test.charAt(i);
           if ((testChar>=0xDC00) && (testChar<=0xDFFF)){ return false; }
           if ((testChar>=0xD800) && (testChar<=0XDBFF)){
            if (i<(length-1)) {
             int secondChar=test.charAt(++i);
             if ( (secondChar<0xDC00) || (secondChar>0xDFFFF) ) { return false; }

             int newChar=( ((testChar-0xD800) * 0x400) + (secondChar-0xDC00) + 0x10000);

             testChar=newChar;
            }
        } 
           returnVal=this.isNameStartChar(testChar);
           if (!returnVal) {break;}
          }
        }
      }//end for
    }//end if
  }//end length>0
  return returnVal;
}



/** Checks a String to see if it conforms to the NMTOKEN rules. Works using either XML 1.0 or 1.1, depending on the rules file used to set the ranges. **/
public boolean isNMTOKEN(String test) {
  boolean returnVal=false;
  int length=test.length();

  if (length>0) {
     for (int i=1; i<length; i++) {
        int testChar=test.charAt(i);
        if ((testChar>=0xDC00) && (testChar<=0xDFFF)){ return false; }
        if ((testChar>=0xD800) && (testChar<=0XDBFF)){
         if (i<(length-1)) {

          int secondChar=test.charAt(++i);
          if ( (secondChar<0xDC00) || (secondChar>0xDFFFF) ) { return false; }

          int newChar=( ((testChar-0xD800) * 0x400) + (secondChar-0xDC00) + 0x10000);

          testChar=newChar;
         }
        }
        returnVal=this.isNameChar(testChar);
        if (!returnVal) {break;}
     }//end for
  }//end length>0
  return returnVal;
}

/** Checks a String to see if it conforms to the NMTOKENS rules. Works using either XML 1.0 or 1.1, depending on the rules file used to set the ranges. **/

public boolean isNMTOKENS(String test) {
  boolean returnVal=false;
  int length=test.length();

  if (length>0) {
     for (int i=0; i<length; i++) {
        int testChar=test.charAt(i);
        if ((testChar>=0xDC00) && (testChar<=0xDFFF)){ return false; }
        if ((testChar>=0xD800) && (testChar<=0XDBFF)){
         if (i<(length-1)) {
          int secondChar=test.charAt(++i);
          if ( (secondChar<0xDC00) || (secondChar>0xDFFFF) ) { return false; }

          int newChar=( ((testChar-0xD800) * 0x400) + (secondChar-0xDC00) + 0x10000);

          testChar=newChar;
         }
        } 
        returnVal=this.isNameChar(testChar);
        if (!returnVal) {returnVal=this.isSpace(testChar);}
        if (!returnVal) {break;}
     }//end for
  }//end length>0
  return returnVal;
}

/** Checks a String to see if it conforms to the NAMES rules. Works using either XML 1.0 or 1.1, depending on the rules file used to set the ranges. **/
public boolean isNames(String test) {
  boolean returnVal=false;
  boolean mode=true;
//while mode==true, will look for either nameStartChar or Space
//while mode==false, will look for either nameChar or Space
//hitting space->true
//hitting a nameChar when mode is true->false

  int length=test.length();

  if (length>0) {
     for (int i=0; i<length; i++) {
        int testChar=test.charAt(i);
        if ((testChar>=0xDC00) && (testChar<=0xDFFF)){ return false; }
        if ((testChar>=0xD800) && (testChar<=0XDBFF)){
         if (i<(length-1)) {

          int secondChar=test.charAt(++i);
          if ( (secondChar<0xDC00) || (secondChar>0xDFFFF) ) { return false; }

          int newChar=( ((testChar-0xD800) * 0x400) + (secondChar-0xDC00) + 0x10000);

          testChar=newChar;
         }
        }
        if (mode) {
          //mode==true
          returnVal=this.isNameStartChar(testChar);
          if (!returnVal) {
            returnVal=this.isSpace(testChar);
          } else {
            mode=false;
          }//end returnVal
        } else {
        //mode==false
          returnVal=this.isNameChar(testChar);
          if (!returnVal) {
            returnVal=this.isSpace(testChar);
            if (returnVal) {mode=true;}
          }
        }//end mode
        if (!returnVal) {break;}
     }//end for
  }//end length>0


  return returnVal;
}

/** Checks a String to see if it conforms to the Pubid rules. XML 1.1 didn't make any changes to the Pubid rules. **/
public boolean isPubid(String test) {
  boolean returnVal=false;
  int length=test.length();

  if (length>0) {
     for (int i=0; i<length; i++) {
        int testChar=test.charAt(i);
        if ((testChar>=0xDC00) && (testChar<=0xDFFF)){ return false; }
        if ((testChar>=0xD800) && (testChar<=0XDBFF)){
         if (i<(length-1)) {

          int secondChar=test.charAt(++i);
          if ( (secondChar<0xDC00) || (secondChar>0xDFFFF) ) { return false; }

          int newChar=( ((testChar-0xD800) * 0x400) + (secondChar-0xDC00) + 0x10000);

          testChar=newChar;
         }
        } 
        returnVal=this.isPubidChar(testChar);
        if (!returnVal) {break;}
     }//end for
  }//end length>0
  return returnVal;
}



public String toString(){
   StringBuffer returnVal=new StringBuffer();
   returnVal.append("Namespace Aware: "+namespaceAware+"\n");
   returnVal.append("Char: \n");
   if (chars!=null){
   returnVal.append(chars.toString());
   } else {returnVal.append("Null");}
   returnVal.append("\n");

   returnVal.append("NameStartChar: \n");
   if (nameStartChar!=null){
   returnVal.append(nameStartChar.toString());
   } else {returnVal.append("Null");}
   returnVal.append("\n");

   returnVal.append("AddlNameChar: \n");
   if (addlNameChar!=null){
   returnVal.append(addlNameChar.toString());
   } else {returnVal.append("Null");}
   returnVal.append("\n");

   returnVal.append("Space:\n");
   if (space!=null){
   returnVal.append(space.toString());
   } else {returnVal.append("Null");}
   returnVal.append("\n");

   returnVal.append("BaseChar:\n");
   if (baseChar!=null){
   returnVal.append(baseChar.toString());
   } else {returnVal.append("Null");}
   returnVal.append("\n");

   returnVal.append("Ideographic:\n");
   if (ideographic!=null){
   returnVal.append(ideographic.toString());
   } else {returnVal.append("Null");}
   returnVal.append("\n");

   returnVal.append("CombiningChar:\n");
   if (combiningChar!=null){
   returnVal.append(combiningChar.toString());
   } else {returnVal.append("Null");}
   returnVal.append("\n");

   returnVal.append("Digit:\n");
   if (digit!=null){
   returnVal.append(digit.toString());
   } else {returnVal.append("Null");}
   returnVal.append("\n");

   returnVal.append("Extender:\n");
   if (extender!=null){
   returnVal.append(extender.toString());
   } else {returnVal.append("Null");}
   returnVal.append("\n");

   returnVal.append("PubidChar:\n");
   if (pubidChar!=null){
   returnVal.append(pubidChar.toString());
   } else {returnVal.append("Null");}
   returnVal.append("\n");

   
   return returnVal.toString();
}

public void print(){
   System.out.println(this.toString());
}

}