package com.simonstl.gorille;

/*
The contents of this file are subject to the Mozilla Public License  Version 1.1 (the "License"); you may not use this file except in  compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/
Software distributed under the License is distributed on an "AS IS"  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  License for the specific language governing rights and limitations under  the License.

The Original Code is available at  http://simonstl.com/projects/fragment/original.

The Initial Developer of the Original Code is Simon St.Laurent. Portions created by Simon St.Laurent are Copyright (C) 2002. All Rights Reserved.

Contributor(s):
*/

/**
 *
 * <p>The docProcI interface provides access to low-level information in an XML
 * (or other markup, since this processing works below even the
 * well-formedness processing in an XML parser.  All information is passed
 * as StringBuffer objects and returned as (possibly changed) StringBuffer
 * objects.  The context object provides the element/attribute context as well
 * as the current namespace declaration stack. </p>
 *
 *
 * <p>One issue in particular is worth noting.  Namespace declarations may not appear until
 * after the first use of an element in that namespace.  This is an unfortunate side-effect
 * of the attribute-based syntax used by Namespaces in XML.  A look-ahead strategy and two
 * context trees (one reflecting the initial document, one reflecting the result) may
 * correct this in the future.</p>
 *
 * @author  simonstl
 */

public interface DocProcI {

    /**
     * setContext() connects this processor to a particular context object
     * coming from the parser.
     */
    public void setContext(ContextI context);

    /**
     * getContext() is useful if you need to chain one processor to another
     * and connect the contexts.
     */
    public ContextI getContext();


    /**
     * setChainListener() lets you set a processor which will receive the
     * events from this processor before the information is returned to the
     * original filter.
     */
    public void setChainListener(DocProcI listener);

    /**
     * getChainListener() lets you find out which object is getting events
     * from this processor.
     */
    public DocProcI getChainListener();

//OTag - open; ETag - empty; CTag - close

    /**
     *
     * Reports the XML declaration, if one is present.
     *
     * The StringBuffer content will contain the entire XML declaration.
     */
    public StringBuffer XMLdecl (StringBuffer content) throws GorilleException;

    /**
     *
     * Reports the DOCTYPE declaration, if one is present..
     *
     * The StringBuffer content will contain the entire DOCTYPE declaration.
     */
    public StringBuffer DOCTYPE (StringBuffer content) throws GorilleException;


    /**
     *
     * Indicates that the parser has begun reading an open or empty tag.
     *
     * The StringBuffer content will always contain "<" for this event
     * if an XML document is the source of the event.
     */
    public StringBuffer startElementOTag (StringBuffer content) throws GorilleException;


    /**
     *
     * Indicates that the parser has begun reading a closing tag.
     *
     * The StringBuffer content will always contain "</" for this event
     * if an XML document is the source of the event.
     */
    public StringBuffer startElementCTag (StringBuffer content) throws GorilleException;


    /**
     *
     * Reports the name of the element described by the current tag.
     *
     * The StringBuffer content will always contain the name of the element.
     */

    public StringBuffer elementName (StringBuffer content) throws GorilleException;



    /**
     *
     * Reports whitespace inside of start, end, and empty tags, except whitespace contained
     * in attribute values.
     *
     * Note that element names are reported as characters (and possibly, outside of XML,
     * entities, etc.)  Collect these events and process in the tagSpace event.
     *
     * The StringBuffer content will contain all the space in any given whitespace area.
     */
    public StringBuffer tagSpace (StringBuffer content) throws GorilleException;

    /**
     * Reports the name of an attribute described by the current tag.
     *
     * The StringBuffer content will always contain the name of the attribute.
     */
    public StringBuffer attName (StringBuffer content) throws GorilleException;





    /**
     *
     * Reports when the equals sign in an attribute name/value pair is encountered.
     *
     * attribute names are reported as characters (and possibly,
     * outside of XML, entities, etc.)
     * Collect these events and process in the attEquals event.
     * The StringBuffer content will always contain "="; whitespace before or after
     * the = will be reported using the tagSpace event.
     */
    public StringBuffer attEquals (StringBuffer content) throws GorilleException;

    /**
     *
     * Reports the start quote type of an attribute described by the current tag.
     *
     * The StringBuffer content will always contain either " or '.
     */
    public StringBuffer attStartQuote (StringBuffer content) throws GorilleException;

    /**
     *
     * Reports the end quote type of an attribute described by the current tag.
     * The end quote will always be the same as the start quote; this is largely a convenience
     * method for developers who don't want to spend their time counting quotes.
     *
     * The StringBuffer content will always contain either " or '.
     */
    public StringBuffer attEndQuote (StringBuffer content) throws GorilleException;

    /**
     *
     * Indicates that the parser has ended reading an open tag.
     *
     * The StringBuffer content will always contain ">" for this event
     * if an XML document is the source of the event.
     */
    public StringBuffer endElementOTag (StringBuffer content) throws GorilleException;

    /**
     *
     * Indicates that the parser has ended reading an empty tag.
     *
     * The StringBuffer content will always contain "/>" for this event
     * if an XML document is the source of the event.
     */

    public StringBuffer endElementETag (StringBuffer content) throws GorilleException;

    /**
     *
     * Indicates that the parser has ended reading a closing tag.
     *
     * The StringBuffer content will always contain ">" for this event
     * if an XML document is the source of the event.
     */
    public StringBuffer endElementCTag (StringBuffer content) throws GorilleException;


    /**
     *
     * Reports raw textual content contained by elements or attributes.
     *
     * The StringBuffer content will contain characters. Parsers may report multiple
     * chars events as necessary, and other events may be interspersed as necessary.
     */
    public StringBuffer chars (StringBuffer content) throws GorilleException;


    /**
     *
     * Reports decimal character references.
     *
     * The StringBuffer content will contain the entire reference - "&#64;" for example.
     */
    public StringBuffer decCharRef (StringBuffer content) throws GorilleException;

    /**
     *
     * Reports hexadecimal character references.
     *
     * The StringBuffer content will contain the entire reference - "&#x64;" for example.
     */
    public StringBuffer hexCharRef (StringBuffer content) throws GorilleException;

    /**
     *
     * Reports entity references.
     *
     * The StringBuffer content will contain the entire reference - "&amp;" for example.
     */
    public StringBuffer entRef (StringBuffer content) throws GorilleException;

    /**
     *
     * Reports the beginning of comments.
     *
     * The StringBuffer content will contain the opening "<!--"
     */
    public StringBuffer commentStart (StringBuffer content) throws GorilleException;
/**
     *
     * Reports comment contents.
     *
     * The StringBuffer content will contain the contents of the comment.
 * There may be multiple commentContent events if the content is long.
     */
    public StringBuffer commentContent (StringBuffer content) throws GorilleException;

    /**
     *
     * Reports comments.
     *
     * The StringBuffer content will contain the closing "-->"
     */
    public StringBuffer commentEnd (StringBuffer content) throws GorilleException;


    /**
     *
     * Reports that a processing instruction has started.
     *
     * The StringBuffer content will contain the opening - "<?".
     */
    public StringBuffer PIStart (StringBuffer content) throws GorilleException;

    /**
     *
     * Reports a processing instruction target.
     *
     * The StringBuffer content will contain the target - "xml-stylesheet", for example.
     */
    public StringBuffer PITarget (StringBuffer content) throws GorilleException;

    /**
     *
     * Reports the whitespace between a processing instruction's target and its data.
     *
     * The StringBuffer content will contain just whitespace.
     */
    public StringBuffer PISpace (StringBuffer content) throws GorilleException;

    /**
     *
     * Reports the data in a processing instruction.
     *
     * The StringBuffer content will contain all of the data in one (potentially large) piece.
     */
    public StringBuffer PIData (StringBuffer content) throws GorilleException;

    /**
     *
     * Reports that a processing instruction has ended.
     *
     * The StringBuffer content will contain the opening - "?>".
     */
    public StringBuffer PIEnd (StringBuffer content) throws GorilleException;

    /**
     *
     * Reports that a CDATA section has started.
     *
     * The StringBuffer content will contain the opening - "<[CDATA[".
     */
    public StringBuffer CDATAStart (StringBuffer content) throws GorilleException;

    /**
     *
     * Reports that a CDATA section has ended.
     *
     * The StringBuffer content will contain the opening - "]]>".
     */
    public StringBuffer CDATAEnd (StringBuffer content) throws GorilleException;

    /** A conduit for warning messages */
    public GorilleException warning (GorilleException e) throws GorilleException;
    /** A conduit for error messages */

    public GorilleException error (GorilleException e) throws GorilleException;
    /** A conduit for fatal error messages */

    public GorilleException fatalError (GorilleException e) throws GorilleException;

}