package com.simonstl.gorille;

/*
The contents of this file are subject to the Mozilla Public License  Version 1.1 (the "License"); you may not use this file except in  compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/
Software distributed under the License is distributed on an "AS IS"  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the  License for the specific language governing rights and limitations under  the License.

The Original Code is available at  http://simonstl.com/projects/fragment/original.

The Initial Developer of the Original Code is Simon St.Laurent. Portions created by Simon St.Laurent are Copyright (C) 2001. All Rights Reserved.

Contributor(s):
*/

import com.simonstl.common.*;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Vector;


/**
 *
 * A context-aware processor for manipulating markup labels and content.
 *
 * Note that this processor does no DOCTYPE processing.
 *
 * This code is based on the parser from the Tiny API for Markup (TAM), but 
 * reports a text version of the document with context rather than abstracting
 * it so heavily to nodes.    For more detail on the events that Ripper produces,
 * see the {@link  com.simonstl.gorille.DocProcI DocProcI} class.
 *
 * This class is not particularly elegant, using methods to isolate code for editorial
 * convenience but relying on variables which are global to the class for most
 * communications inside of the parser.  This will improve with time, hopefully.
 *
 * The easiest way to use this class is to create a Ripper object and assign a handler
 * (and possibly a context, if you want to use one other than the default), then call the
 * parse() method with an appropriate InputStream.
 *
 * This parser does not presently understand character encodings, and in fact regards
 * the XML declaration as a processing instruction.  This will also improve with time.
 *
 * @author  simonstl@simonstl.com
 */
public class Ripper {


    private static final int UNKNOWN=0;

    private static final int ELEMENT_START=1;
    private static final int ELEMENT_END=2;
    private static final int ELEMENT_SPACE=3;

    private static final int ATT_NAME=4;
    private static final int ATT_VALUE=5;
    private static final int ATT_EQUALS=6;

    private static final int PI_TARGET=8;
    private static final int PI_SPACE=9;
    private static final int PI_DATA=10;
    private static final int PI_QUESTION=11;

    private static final int CDATA_START=16;
    private static final int CDATA_CONTENT=17;
    private static final int CDATA_SQUARE=18;
    private static final int CDATA_SQUARE_SQUARE=19;

    private static final int DOCTYPE=32;
    private static final int DOCTYPE_END_SQUARE=33;

    private static final int LESSTHAN=64;
    private static final int LESSTHAN_BANG=65;


    private static final int COMMENT_TEXT=128;
    private static final int COMMENT_FIRST=129;
    private static final int COMMENT_DASH=130;
    private static final int COMMENT_DASH_DASH=131;

    private static final int AMP=256;
    private static final int AMP_SHARP=257;
    private static final int HEX_REF=259;
    private static final int DEC_REF=260;
    private static final int ENTITY=264;

    private static final int START_TAG_COLLECTING=512;

    private static final int XML_DECL_CONTENT=1024;

    private static final int XML_DECL_QUESTION=1025;



    private static final String doubleQuote=new String("\"");
    private static final String singleQuote=new String("\'");
    private static final String openAngle=new String("<");
    private static final String closeAngle=new String(">");
    private static final String closeEmpty=new String("/>");
    private static final String openEndTag=new String("</");
    private static final String openPI=new String("<?");
    private static final String closePI=new String("?>");
    private static final String openComment=new String("<!--");
    private static final String closeComment=new String("-->");
    private static final String equals=new String("=");
    private static final String CDATAstart=new String("<![CDATA[");
    private static final String CDATAend=new String("]]>");

    boolean quoteType=true; //true = "", false = '

    DocProcI handler=null;
    ContextI context=new Context();

    int size=1024;
    //if reading is true, then we haven't reached the end of the buffer
    boolean reading=true;

    //if halt is true we've hit a serious error
    boolean halt=false;

    //if skipChar is true, we skip the rest of processing for this character
    boolean skipChar=false;

    //if targetSpace is true, we've encountered the first space in a PI
    boolean targetSpace=false;

    //whether or not we've finished the element name
    boolean elementNameEstablished=false;

    //whether or not we've already hit a colon in the current name
    boolean colonInName=false;

    //the main buffer for reading
    char[] buf=new char[size];

    //the internal storage buffer for content
    StringBuffer current=new StringBuffer();

    //the internal storage buffer for content
    StringBuffer startTag=new StringBuffer();

    //NodeType for where we were before entitiy
    int beforeEntity=UNKNOWN;

    //if putInBuffer is false, the current character doesn't go in the buffer
    boolean putInBuffer=true;

    //if pastXMLDecl is true, we've encountered something beyond where the XML decl is allowed
    boolean pastXMLDecl=false;

    //if isEmpty is true, it's an empty element
    boolean isEmpty=false;

    //temporary holder for attribute prefix info
    String prefix="";

    //temporary holder for attribute value
    StringBuffer attValue=new StringBuffer();

    //temporary holder for attributes
    protected ArrayList attributes=new ArrayList();

    //temporary holder for element info
    StackableComponent element=new StackableComponent(NodeType.ELEMENT);

    //temporary holder for attribute info
    StackableComponent attribute=new StackableComponent(NodeType.ATTRIBUTE);

    //prior elements
    Vector elementStack=new Vector();

    //PI target for error reporting
    String target="";

    //limit on size of diagnostic content stored in context, in characters
    int limit=16;

    //line-end normalization
    boolean xD=false;

    //if in the DOCTYPE, are we in the internal subset?
    boolean internalSubset=false;

    //procPI used when PI_QUESTION not end of PI
    boolean procPI=true;

    //used to simplify code for separating XML decl, PI
    boolean openPIreported=false;

    //state controls which rules apply where inside of startTags
    int internalState=ELEMENT_START;

    /** Creates a new instance of Ripper */
    public Ripper() {

    }

    public void setMaxBuf(int _size){
        if (size>64) {size=_size;}
    }

    public DocProcI getHandler() {
        return handler;
    }

    public void setHandler(DocProcI _handler) {
        handler=_handler;
    }

    public void parse(Reader r) throws GorilleException {

        //if overflow is true, backup buffer has contents

        int bigPosition=0;
        int numChars=0;
        //state controls which rules apply where
        int state=UNKNOWN;
        elementStack.addElement("-");
        while (reading) {
            //fill buffer
            try {
                numChars=r.read(buf, bigPosition, size);

                if (numChars<size) {reading=false;}

                //read through bytes
                //key characters are <, >, &,

                for (int i=0;i<numChars;i++){
                    char currentChar=buf[i];
//System.out.println("State: "+String.valueOf(state)+" char "+String.valueOf(currentChar));
                    putInBuffer=true;

                    if ((currentChar=='&') && (!(state==START_TAG_COLLECTING)))  {
                        state=processAmpersand(state);
                    }
                    if ( (currentChar=='<') &&  (state==UNKNOWN)) {
                        if (current.length()>0) {

                            StackableComponentI text=new StackableComponent();
                            text.setType(NodeType.CHARS);
                            int length=current.length();
                            if (length < (limit-1) ){
                                //nothing
                            } else {
                                length=limit-1;
                            }
                            text.setDiagValue(current.substring(0,length));
                            context.addNode(text);

                            handler.chars(current);
                            current=new StringBuffer();
                            pastXMLDecl=true;
                        }
                        state=LESSTHAN;
                        putInBuffer=false;
                    } else {


//begin handlers for particular states-----------------------------------------------------
                        if ((currentChar=='<') && ((!((state==UNKNOWN) || (state==COMMENT_TEXT) || (state==PI_SPACE) ||
                                (state==PI_DATA) || (state==PI_QUESTION) || (state==CDATA_CONTENT) ||
                                (state==CDATA_SQUARE) || (state==CDATA_SQUARE_SQUARE) ||
                                ((state==DOCTYPE) && internalSubset)))) ) {




                            fatalError("Too many <s encountered "+ String.valueOf(state));
                        }
                        /*if ((state==UNKNOWN) || (state==ATT_VALUE)) {
                        //line-end normalization
                        lineEndNorm(currentChar);
                        }*/

                        state=processEntities(currentChar, state);

                        state=processDOCTYPE(currentChar, state);

                        state=processComments(currentChar, state);

                        state=processCDATA(currentChar, state);


                        if (state==LESSTHAN_BANG){
                            if (currentChar=='-'){
                                state=COMMENT_FIRST;
                                putInBuffer=false;
                            } else if (currentChar=='['){
                                state=CDATA_START;
                                current.append("<![");
                                putInBuffer=false;
                            } else {
                                state=DOCTYPE;
                                current.append('<');
                                current.append('!');
                            }
                        }

                        if (!pastXMLDecl) {
                            //xmlDecl
                            state=processXMLDecl(currentChar, state);
                        } else {
                            state=processPIs(currentChar, state);
                        }
                        state=processEndTag(currentChar, state);

                        state=processStartTagCollecting(currentChar, state);

                        if (state==LESSTHAN){
                            if (currentChar=='?'){
                                state=PI_TARGET;
                                targetSpace=false;
                                openPIreported=false;
                                current=new StringBuffer("<?");
                                putInBuffer=false;
                                //                            current=new StringBuffer();
                            } else if (currentChar=='!') {
                                state=LESSTHAN_BANG;
                                putInBuffer=false;
                                pastXMLDecl=true;
                            } else if (currentChar=='/') {
                                elementNameEstablished=false;
                                state=ELEMENT_END;
                                handler.startElementCTag(new StringBuffer(openEndTag));
                                putInBuffer=false;
                                pastXMLDecl=true;
                            } else if (context.isSpace(currentChar)) {
                                //whitespace - add LF, CR
                                fatalError("Whitespace at wrong place");
                            } else {
                                state=START_TAG_COLLECTING;
                                internalState=ELEMENT_START;
                                handler.startElementOTag(new StringBuffer(openAngle));
                                elementNameEstablished=false;
                                processStartTagCollecting(currentChar, state);
                                //setup for element
                                putInBuffer=true;
                                pastXMLDecl=true;
                            }
                        }



                    } //end if char=<
                    if (putInBuffer){
                        current.append(currentChar);
                    }
//processState("End  ", state, i, currentChar, putInBuffer);
                    if (halt) {break;}
                } //end for

                //handle end-of-buffer situation

            } catch (Exception e) {
                System.out.println("blam!" + e);
                reading=false;
            }



        }//end while (reading)
    }

    protected int processAmpersand(int state) throws GorilleException {
//System.out.println ("Hit an ampersand in state "+String.valueOf(state));

        if ((state==UNKNOWN) || (state==ATT_VALUE)) {
            beforeEntity=state;
            if (current.length()>0) {
                if (state==UNKNOWN) {
                    StackableComponentI text=new StackableComponent();
                    text.setType(NodeType.CHARS);
                    int length=current.length();
                    if (length < (limit-1) ){
                        //nothing
                    } else {
                        length=limit-1;
                    }
                    text.setDiagValue(current.substring(0,length));
                    context.addNode(text);
                }

                handler.chars(current);
                current=new StringBuffer();
            }
            state=AMP;
            putInBuffer=false;
            skipChar=true;
        } else if
        ((state==COMMENT_TEXT) || (state==COMMENT_DASH) ||
                ((state==DOCTYPE) && internalSubset) ||
                (state==PI_TARGET) || (state==PI_SPACE) ||
                (state==PI_DATA) || (state==CDATA_START) ||
                (state==CDATA_CONTENT) || (state==CDATA_SQUARE) ||
                (state==CDATA_SQUARE_SQUARE)
        ) {
            //everything's okay, so let it go through
        } else {
            //yecch... must be someplace illegal.
            fatalError("Inappropriate ampersand in state "+ String.valueOf(state));
        }
        return state;
    }

    protected int internalProcessAmpersand(int state) throws GorilleException {
//System.out.println ("Hit an ampersand in state "+String.valueOf(state));

        if ((state==UNKNOWN) || (state==ATT_VALUE)) {
            beforeEntity=state;


            current=new StringBuffer();

            state=AMP;
            putInBuffer=false;
            skipChar=true;
        } else if
        ((state==COMMENT_TEXT) || (state==COMMENT_DASH) ||
                ((state==DOCTYPE) && internalSubset) ||
                (state==PI_TARGET) || (state==PI_SPACE) ||
                (state==PI_DATA) || (state==CDATA_START) ||
                (state==CDATA_CONTENT) || (state==CDATA_SQUARE) ||
                (state==CDATA_SQUARE_SQUARE)
        ) {
            //everything's okay, so let it go through
        } else {
            //yecch... must be someplace illegal.
            fatalError("Inappropriate ampersand in state "+ String.valueOf(state));
        }
        return state;
    }


    protected int processDOCTYPE(char currentChar, int state) throws GorilleException {
        if (state==DOCTYPE_END_SQUARE){
            if (currentChar=='>'){
                current.append('>');
                handler.DOCTYPE(current);
                state=UNKNOWN;
                putInBuffer=false;
                current=new StringBuffer();
            }
        }

        if (state==DOCTYPE) {

            //we're only looking for >, [, or ]
            if (!internalSubset){
                if (currentChar=='>'){
                    current.append('>');
                    handler.DOCTYPE(current);
                    state=UNKNOWN;
                    putInBuffer=false;
                    current=new StringBuffer();

                } else if (currentChar=='[') {
                    internalSubset=true;
                }
            } else {
                if (currentChar==']'){
                    state=DOCTYPE_END_SQUARE;
                }
            }
        }
        return state;

    }

    protected int processComments(char currentChar, int state) throws GorilleException {
        if (state==COMMENT_DASH_DASH) {
            if (currentChar=='>'){
                StackableComponentI comment=new StackableComponent();
                comment.setType(NodeType.COMMENT);
                int length=current.length();
                if (length < (limit-1) ){
                    //nothing
                } else {
                    length=limit-1;
                }
                comment.setDiagValue(current.substring(0,length));
                context.addNode(comment);

                handler.commentContent(current);
                handler.commentEnd(new StringBuffer(closeComment));
                state=UNKNOWN;
                current=new StringBuffer();
                putInBuffer=false;
            } else {
                warning("Malformed comment --");
                current.append("--");
                state=COMMENT_TEXT;
            }
        }

        if (state==COMMENT_DASH) {
            if (currentChar=='-'){
                state=COMMENT_DASH_DASH;
                putInBuffer=false;
            } else {
                current.append('-');
                state=COMMENT_TEXT;
            }
        }

        if (state==COMMENT_TEXT) {
            if (currentChar=='-'){
                state=COMMENT_DASH;
                putInBuffer=false;
            }
        }

        if (state==COMMENT_FIRST) {
            if (currentChar=='-'){
                handler.commentStart(new StringBuffer(openComment));
                state=COMMENT_TEXT;
                putInBuffer=false;
            } else {
                state=UNKNOWN;
                fatalError("Botched comment");
            }
        }
        return state;

    }

    protected int processCDATA(char currentChar, int state) throws GorilleException {
        if (state==CDATA_SQUARE_SQUARE){
            if (currentChar=='>'){
                state=UNKNOWN;

                StackableComponentI CDATA=new StackableComponent();
                CDATA.setType(NodeType.CDATA);
                int length=current.length();
                if (length < (limit-1) ){
                    //nothing
                } else {
                    length=limit-1;
                }
                CDATA.setDiagValue(current.substring(0,length));
                context.addNode(CDATA);

                handler.chars(current);
                handler.CDATAEnd(new StringBuffer(CDATAend));
                putInBuffer=false;
                current=new StringBuffer();
            } else {
                //false alarm
                current.append("]]");
                state=CDATA_CONTENT;
            }
        }

        if (state==CDATA_SQUARE){
            if (currentChar==']'){
                state=CDATA_SQUARE_SQUARE;
                putInBuffer=false;
            } else {
                //false alarm
                current.append(']');
                state=CDATA_CONTENT;
            }
        }
        if (state==CDATA_CONTENT){
            if (currentChar==']'){
                state=CDATA_SQUARE;
                putInBuffer=false;
            }
        }

        if (state==CDATA_START) {
            if (current.length()==9) {
                if (current.toString().equals("<![CDATA[")) {
                    handler.CDATAStart(new StringBuffer(CDATAstart));
                    current=new StringBuffer();
                    state=CDATA_CONTENT;
                }   else {
                    state=UNKNOWN;
                    fatalError("Botched CDATA section: "+ current.toString());
                }
            }
        }
        return state;

    }

    protected int processXMLDecl(char currentChar, int state) throws GorilleException{
        if (state==XML_DECL_QUESTION){
//System.out.println("Current char:"+String.valueOf(currentChar));
            if (currentChar=='>'){

                //end XML decl
                current.append('>');
                handler.XMLdecl(current);
                boolean acceptable=context.setXMLDeclaration(current.toString());
                if (!(acceptable)) {fatalError("bad XML declaration: "+current);}
                state=UNKNOWN;
                putInBuffer=false;
                current=new StringBuffer();
                pastXMLDecl=true;

            } else {
                fatalError("bad XML declaration: "+current);

            }
        }


        if (state==XML_DECL_CONTENT) {
            if (currentChar=='?'){
                state=XML_DECL_QUESTION;
            }
        }

        if (state==PI_TARGET){

            if (context.isSpace(currentChar)){
                state=PI_SPACE;
                putInBuffer=true;
                target=current.substring(2,current.length());

                if (target.equals("xml")){
                    state=XML_DECL_CONTENT;
                } else {
                    handler.PIStart(new StringBuffer(openPI));
                    handler.PITarget(new StringBuffer(target));
                    current=new StringBuffer();
                    targetSpace=true;
                    pastXMLDecl=true;
                }
            } else if (currentChar=='?'){
                target=current.substring(2,current.length());
                if (target.equals("xml")){
                    state=XML_DECL_QUESTION;
                } else {
                    handler.PIStart(new StringBuffer(openPI));
                    handler.PITarget(current);
                    state=PI_QUESTION;
                    putInBuffer=false;

                    current=new StringBuffer();
                }
            }
        }

        return state;
    }

    protected int processPIs(char currentChar, int state) throws GorilleException {
        if (state==PI_QUESTION){
//System.out.println("Current char:"+String.valueOf(currentChar));
            if (currentChar=='>'){
                StackableComponentI procInst=new StackableComponent();
                procInst.setType(NodeType.PI);
                procInst.setLocalName(target);
                int length=current.length();
                if (length < (limit-1) ){
                    //nothing
                } else {
                    length=limit-1;
                }
                procInst.setDiagValue(current.substring(0,length));
                context.addNode(procInst);
                handler.PIData(current);
                handler.PIEnd(new StringBuffer(closePI));
                state=UNKNOWN;
                putInBuffer=false;
                current=new StringBuffer();
            } else if (targetSpace) {
                current.append('?');
                state=PI_DATA;
                procPI=false;
            } else {
                fatalError("Faulty PI with target: "+target);

            }
        }

        if (state==PI_DATA) {
//System.out.println("ProcPI is "+procPI);
//System.out.println("Buffer contains: "+current.toString());
            if (procPI) {
                if (currentChar=='?'){
                    putInBuffer=false;
                    state=PI_QUESTION;
                }
            } else {
                procPI=true;
            }
        }

        if (state==PI_SPACE) {
            if (context.isSpace(currentChar)){
                putInBuffer=true;
            } else {
                handler.PISpace(current);
                current=new StringBuffer();
                state=PI_DATA;
            }
        }

        if (state==PI_TARGET){
            if (openPIreported){

            } else {
                handler.PIStart(new StringBuffer(openPI));
                openPIreported=true;
            }
            if (context.isSpace(currentChar)){
                state=PI_SPACE;
                putInBuffer=true;
                target=current.substring(2,current.length());
                handler.PITarget(new StringBuffer(target));
                current=new StringBuffer();
                targetSpace=true;
            } else if (currentChar=='?'){
                handler.PIStart(new StringBuffer(openPI));
                target=current.substring(2,current.length());
                handler.PITarget(new StringBuffer(target));
                state=PI_QUESTION;
                putInBuffer=false;

                current=new StringBuffer();

            }
        }
        return state;

    }

    protected int processEndTag (char currentChar, int state) throws GorilleException {
        if (state==ELEMENT_END){

            //check for space or >, otherwise report end of element
            if (currentChar=='>') {
                //report endElement - strip space from end of element if nec
                putInBuffer=false;
                handler.elementName(current);
                String name=current.toString();

                String startName=(String) elementStack.lastElement();

                if (! (name.equals(startName))){
                    fatalError(name+" does not match "+startName);
                }

                int size=elementStack.size() -1;
//System.out.println(size);
                elementStack.removeElementAt(size);

                element.setPrefix(prefix);
                //URI will be set when element reported
                element.setLocalName(name);
                handler.endElementCTag(new StringBuffer(closeAngle));
                state=UNKNOWN;
                colonInName=false;
                prefix="";
                context.endChild();
                current=new StringBuffer();
            } else if (currentChar==':') {
                if (colonInName){
                    fatalError("Too many colons in name");
                } else {

                    colonInName=true;
                }
            }

        }

        return state;



    }

    protected int processStartTagCollecting(char currentChar, int state) throws GorilleException {

        if (state==START_TAG_COLLECTING){
//System.out.println("State:           "+String.valueOf(state));
//System.out.println("Char:            "+String.valueOf(currentChar));
//System.out.println("Internal State 1:"+String.valueOf(internalState));
            startTag.append(currentChar);
            processStartTagForContext(currentChar);
//System.out.println("Internal State 2:"+String.valueOf(internalState));
            if (internalState==UNKNOWN){
                state=UNKNOWN;
                internalState=ELEMENT_START;
                isEmpty=false;
                //report the whole thing
                int bufferSize=startTag.length();
                for (int i=0; i<bufferSize; i++){
                    char bufferChar=startTag.charAt(i);
                    putInBuffer=true;
                    processStartTagForReporting(bufferChar);
                    if (putInBuffer){
                        current.append(bufferChar);
                    }

                    if (halt) {break;}
                }
                internalState=ELEMENT_START;
                startTag=new StringBuffer();
                isEmpty=false;
            }
        }
        return state;

    }

    protected void processStartTagForContext(char currentChar) throws GorilleException {



        if (currentChar=='&'){
            if (internalState==ATT_VALUE) {
                attValue.append(current);
            }
            internalState=internalProcessAmpersand(internalState);

        }

        internalState=internalProcessEntities(currentChar, internalState);

        if (internalState==ATT_VALUE){
            if (( (currentChar=='"') && (quoteType) ) ||
                    ( (currentChar=='\'') && (!(quoteType)) )) {
                internalState=ELEMENT_SPACE;
                putInBuffer=false;

                attValue.append(current);

                String value=attValue.toString();
                if (attribute.getPrefix().equals("xmlns")){
                    context.declarePrefix(attribute.getLocalName(), value);
                } else if ( (attribute.getPrefix().equals("")) && (attribute.getLocalName().equals("xmlns")) ){
                    context.declarePrefix("", value);
                }
                int length=current.length();
                if (length < (limit-1) ){
                    //nothing
                } else {
                    length=limit-1;
                }
                attribute.setDiagValue(current.substring(0,length));

                attributes.add(attribute);
                attribute=new StackableComponent();
                attValue=new StringBuffer();
                current=new StringBuffer();
                skipChar=true;
            }



        }

        if (internalState==ATT_EQUALS){
            if (context.isSpace(currentChar)){
                putInBuffer=false;
            } else if (currentChar=='"') {
                putInBuffer=false;
                quoteType=true;
                internalState=ATT_VALUE;
            } else if (currentChar=='\'') {
                putInBuffer=false;
                quoteType=false;
                internalState=ATT_VALUE;
            }

        }


        if (internalState==ATT_NAME){
            if (currentChar=='='){
                internalState=ATT_EQUALS;
                putInBuffer=false;

                String name=current.toString();
                attribute.setPrefix(prefix);
                //URI will be set when element reported
                attribute.setLocalName(name);
                prefix="";
                current=new StringBuffer();
                colonInName=false;

            } else if (currentChar==':') {
                if (colonInName){
                    fatalError("Too many colons in name");
                } else {
                    prefix=current.toString();
                    colonInName=true;
                    current=new StringBuffer();
                    putInBuffer=false;

                }
            } else if (context.isSpace(currentChar)){
                putInBuffer=false;
            }
        }


        if (internalState==ELEMENT_SPACE){
            if (!skipChar) {
                //check for space or >, otherwise start attribute
                if (context.isSpace(currentChar)) {
                    putInBuffer=false;
                } else if (currentChar=='>') {
                    putInBuffer=false;
                    if (!elementNameEstablished) {
                        String name=current.toString();
                        element.setPrefix(prefix);
                        //URI will be set when element reported
                        element.setLocalName(name);
                    }
                    startTagDone();
                    internalState=UNKNOWN;
                } else if (currentChar=='/') {
                    isEmpty=true;
                    putInBuffer=false;

                } else {
                    current=new StringBuffer();
                    internalState=ATT_NAME;
                }
            } else {
                skipChar=false;
            }
        }

        if (internalState==ELEMENT_START){
            //check for whitespace, save
            if (context.isSpace(currentChar)){
                internalState=ELEMENT_SPACE;

                String name=current.toString();
                element.setPrefix(prefix);
                //URI will be set when element reported
                element.setLocalName(name);
                elementNameEstablished=true;
                colonInName=false;
                prefix="";
                putInBuffer=false;
            } else if (currentChar==':') {
                if (colonInName){
                    fatalError("Too many colons in name");
                } else {
                    prefix=current.toString();
                    colonInName=true;
                    current=new StringBuffer();
                    putInBuffer=false;
                }

            } else if (currentChar=='/') {
                //empty element
                isEmpty=true;
                putInBuffer=false;

            } else if (currentChar=='>') {
                //element name=buffer
                putInBuffer=false;

                String name=current.toString();
                element.setPrefix(prefix);
                //URI will be set when element reported
                element.setLocalName(name);

                elementNameEstablished=true;
                colonInName=false;
                prefix="";
                startTagDone();
                internalState=UNKNOWN;
            }
        }


    }

    protected void startTagDone(){

        element.setUri(context.getUri(element.getPrefix()));
        int numAtts=attributes.size();
        for (int i=0; i<numAtts; i++) {
            StackableComponent thisOne=(StackableComponent) attributes.get(i);
            String pref=thisOne.getPrefix();
            if (!(pref.equals(""))) {
                thisOne.setUri(context.getUri(pref));
            }
            element.addAtt(thisOne);
        }

        context.startChild(element);


        if (isEmpty) {
            context.endChild();
        }   else {
            elementStack.addElement(element.getQName());
        }

        element=new StackableComponent();
        attributes.clear();
        current=new StringBuffer();
    }

    protected void processStartTagForReporting(char currentChar) throws GorilleException {
        //System.out.println(current);

        if (currentChar=='&'){
            internalState=processAmpersand(internalState);
        }

        internalState=processEntities(currentChar, internalState);

        if (internalState==ATT_VALUE){
            if (( (currentChar=='"') && (quoteType) ) ||
                    ( (currentChar=='\'') && (!(quoteType)) )) {
                internalState=ELEMENT_SPACE;
                putInBuffer=false;
                handler.chars(current);


                if (currentChar=='"') {
                    handler.attEndQuote(new StringBuffer(doubleQuote));
                } else {
                    handler.attEndQuote(new StringBuffer(singleQuote));
                }
                current=new StringBuffer();
                skipChar=true;
            }


        }

        if (internalState==ATT_EQUALS){
            if (context.isSpace(currentChar)){
                putInBuffer=true;
            } else if (currentChar=='"') {
                putInBuffer=false;
                if (current.length() > 0){
                    handler.tagSpace(current);
                    current=new StringBuffer();
                }
                current=new StringBuffer();
                handler.attStartQuote(new StringBuffer(doubleQuote));
                quoteType=true;
                internalState=ATT_VALUE;
            } else if (currentChar=='\'') {
                putInBuffer=false;
                if (current.length() > 0){
                    handler.tagSpace(current);
                    current=new StringBuffer();
                }
                quoteType=false;
                handler.attStartQuote(new StringBuffer(singleQuote));
                internalState=ATT_VALUE;
            }

        }


        if (internalState==ATT_NAME){
            if (currentChar=='='){
                internalState=ATT_EQUALS;
                putInBuffer=false;
                handler.attName(current);
                handler.attEquals(new StringBuffer(equals));
                current=new StringBuffer();
                colonInName=false;

            } else if (currentChar==':') {
                if (colonInName){
                    fatalError("Too many colons in name");
                } else {

                    colonInName=true;

                }
            } else if (context.isSpace(currentChar)){
                putInBuffer=false;
            }
        }


        if (internalState==ELEMENT_SPACE){
            if (!skipChar) {
                //check for space or >, otherwise start attribute
                if (context.isSpace(currentChar)) {
                    putInBuffer=true;
                } else if (currentChar=='>') {
                    putInBuffer=false;
                    if (!elementNameEstablished) {

                    }
                    if (current.length() > 0){
                        handler.tagSpace(current);
                        current=new StringBuffer();
                    }

                    if (isEmpty){
                        handler.endElementETag(new StringBuffer(closeEmpty));
                    } else {

                        handler.endElementOTag(new StringBuffer(closeAngle));
                    }
                    internalState=UNKNOWN;
                } else if (currentChar=='/') {
                    //empty element
                    if (isEmpty){
                        fatalError("too many slashes in empty tag");
                    } else {
                        isEmpty=true;
                        putInBuffer=false;

                    }
                    //element.setIsEmpty(true);
                } else {
                    if (current.length() > 0){
                        handler.tagSpace(current);

                    }
                    current=new StringBuffer();
                    internalState=ATT_NAME;
                }
            } else {
                skipChar=false;
            }
        }

        if (internalState==ELEMENT_START){
            //check for whitespace, save
            if (context.isSpace(currentChar)){
                internalState=ELEMENT_SPACE;
                elementNameEstablished=true;
                colonInName=false;

                handler.elementName(current);
                current=new StringBuffer();
                putInBuffer=true;
            } else if (currentChar==':') {
                if (colonInName){
                    fatalError("Too many colons in name");
                } else {
                    colonInName=true;
                }

            } else if (currentChar=='/') {
                //empty element
                if (isEmpty){
                    fatalError("too many slashes in empty tag");
                } else {
                    isEmpty=true;
                    putInBuffer=false;

                }
                //element.setIsEmpty(true);
            } else if (currentChar=='>') {
                //element name=buffer
                putInBuffer=false;
                handler.elementName(current);

                if (isEmpty){
                    handler.endElementETag(new StringBuffer(closeEmpty));
                } else {

                    handler.endElementOTag(new StringBuffer(closeAngle));
                }
                elementNameEstablished=true;
                current=new StringBuffer();
                colonInName=false;
                prefix="";
                internalState=UNKNOWN;
            }
        }

    }

    protected int processEntities(char currentChar, int state) throws GorilleException {
        if (state==ENTITY) {
            if (currentChar==';'){
                current.append(';');


                handler.entRef(current);


                state=beforeEntity;
                current=new StringBuffer();
            } else {
                current.append(currentChar);
            }
            putInBuffer=false;
        }

        if ((state==HEX_REF) || (state==DEC_REF)){

            if (currentChar==';'){

                current.append(';');

                if (state==HEX_REF){
                    handler.hexCharRef(current);
                } else {
                    handler.decCharRef(current);
                }
                state=beforeEntity;


                putInBuffer=false;
                current=new StringBuffer();
            } else {
                current.append(currentChar);
                putInBuffer=false;

            }

        }

        if (state==AMP_SHARP){
            current.append(currentChar);
            if (currentChar=='x'){

                state=HEX_REF;

            } else {
                state=DEC_REF;

            }

            putInBuffer=false;
        }

        if (state==AMP){

            if(!skipChar){
                if (currentChar=='#'){
                    current.append("&#");
                    state=AMP_SHARP;
                } else {
                    //add check for acceptable chars - someday
                    state=ENTITY;
                    current.append('&');
                    current.append(currentChar);
                }


                putInBuffer=false;
            } else {skipChar=false;}
        }
        return state;
    }

    protected int internalProcessEntities(char currentChar, int state) throws GorilleException {
        if ((state==ENTITY) || (state==HEX_REF) || (state==DEC_REF)) {
//System.out.println(currentChar+"  "+state);
            if (currentChar==';'){
                String entName=current.toString();
                current=new StringBuffer();

                String entValue=context.resolveEntity(entName);
                if (entValue==null){
                    current.append('&');
                    current.append(entName);
                    current.append(';');
                    //System.out.println("HEY! "+entValue);
                } else {
                    current.append(entValue);
                    // System.out.println("HEY! "+entValue);
                }


                state=beforeEntity;
                if (state==ATT_VALUE){
                    attValue.append(current);
                }
                current=new StringBuffer();
                putInBuffer=false;
            }
        }

        if (state==AMP_SHARP){
            current.append(currentChar);
            if (currentChar=='x'){

                state=HEX_REF;

            } else {
                state=DEC_REF;

            }

            putInBuffer=false;
        }

        if (state==AMP){

            if(!skipChar){
                if (currentChar=='#'){
                    current.append("#");
                    state=AMP_SHARP;
                } else {
                    //add check for acceptable chars - someday
                    state=ENTITY;
                    // current.append('&');
                    current.append(currentChar);
                }


                putInBuffer=false;
            } else {skipChar=false;}
        }
        return state;
    }


    protected void lineEndNorm(char currentChar, int state) {
        if ((currentChar==13) || (currentChar==10)){
            if ((currentChar==13) && (!xD)){
                xD=true;
                putInBuffer=false;
            } else if (currentChar==13){
                current.append((char)10);
                putInBuffer=false;
            } else if (xD) {
                current.append((char)10);
            } else if (currentChar==10) {
                putInBuffer=true;
                xD=false;
            }
        }
    }





    /**
     *
     * fatal errors are reported and processing halts
     */
    protected void fatalError(String s) throws GorilleException{
        GorilleException e=new GorilleException(s);
        try {
            handler.fatalError(e);
        } catch (GorilleException ex) {
            //oy, we're hosed
            throw e;
        }
        halt=true;
        reading=false;
    }

    /**
     *
     *  errors are reported and processing halts
     */

    protected void error(String s) throws GorilleException{
        GorilleException e=new GorilleException(s);
        try {
            handler.error(e);
        } catch (GorilleException ex) {
            //oy, we're hosed
            throw e;
        }
        reading=false;
    }

    /**
     *
     *warnings are reported and processing continues
     *
     * */
    protected void warning(String s)  throws GorilleException{
        GorilleException e=new GorilleException(s);
        try {
            handler.warning(e);
        } catch (GorilleException ex) {
            //oy, we're hosed
            throw e;
        }
    }


    public void setFeature(String name, boolean value) {
    }

    public boolean getFeature(String name) {
        return false;
    }

    public void setContext(ContextI context) {
        this.context=context;
    }

    public ContextI getContext() {
        return context;
    }

    public void setDiagLimit (int limit) {
        this.limit=limit;
    }

    public int getDiagLimit() {
        return limit;
    }
}
